/* Turtle.java -- Graphics Turtle.
**
** Copyright (C) 1997-2002 Eric Laroche.  All rights reserved.
**
** @author Eric Laroche <laroche@lrdev.com>
** @version @(#)$Id: Turtle.java,v 1.1 2002/12/26 17:10:11 laroche Exp $
**
** This program is free software;
** you can redistribute it and/or modify it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
*/

package com.lrdev.turtle;

// note: no wildcard imports

// This Turtle graphics implementation is AWT based.
import java.awt.Graphics;
import java.awt.Rectangle;

/** Turtle: Graphics Turtle.
*
* @author Eric Laroche <laroche@lrdev.com>
* @version @(#)$Id: Turtle.java,v 1.1 2002/12/26 17:10:11 laroche Exp $
* @url http://www.lrdev.com/lr/java/javaturtle.html
*
* <p>
* The Graphics Turtle is a cursor that has a position
* (X and Y coordinates), a heading and a pen up/down state.
* Procedures as forward, left, etc are used as drawing idioms.
*
* <p>
* Turtle Graphics are typically known from the Logo programming
* language.  All Logo Turtle primitives are implemented, including their
* abbreviations.
*
* <p>
* Linewidth, linejoin, linecap, etc are implicit part of
* the used graphics (java.awt.Graphics) context.
* These properties can be changed while using the Turtle.
*
* <p>
* Different heading modes (units of mesurement) are provided:
* RADIANS (circle is 2 pi), DEGREE (circle is 360), GRAD (circle is 400).
*
* <p>
* com.lrdev.turtle.Turtle only offers Turtle methods with double float
* arguments; use com.lrdev.turtle.StringTurtle to allow String arguments
* and certain control structures (e.g. repeat).
*
* <p>
* Most Turtle methods return the Turtle itself, so Turtle calls can be
* stacked: forward(10).left(90);
*
* <p>
* The com.lrdev.turtle.Turtle public API is rather fat (68 methods),
* however many of the methods are abbreviations and other kind of aliases.
* The core Turtle primitives consist of 16 methods.
*/
public class Turtle
{
	// -------- version informations

	/** Version.
	*/
	private final static String m_version =
		"Version: @(#)$Id: Turtle.java,v 1.1 2002/12/26 17:10:11 laroche Exp $";

	/** Author.
	*/
	private final static String m_author =
		"Author: Eric Laroche <laroche@lrdev.com>";

	/** Copyright.
	*/
	private final static String m_copyright =
		"Copyright (C) 1997-2002 Eric Laroche.  All rights reserved.";

	// note: have this disclaimer as one string literal,
	// i.e. no '+' string concatenation (C string concatenation not supported by Java).
	/** Disclaimer.
	*/
	private final static String m_disclaimer =
		"This program is distributed in the hope that it will be useful,\nbut WITHOUT ANY WARRANTY; without even the implied warranty of\nMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.";

	// -------- heading modes / unit of mesurement

	/** Radians heading mode (unit of mesurement).
	*
	* @see #DEGREE
	* @see #GRAD
	*/
	public final static int RADIANS = 1;

	/** Degree heading mode (unit of mesurement).
	*
	* @see #RADIANS
	* @see #GRAD
	*/
	public final static int DEGREE = 2;

	/** Grad heading mode (unit of mesurement).
	*
	* @see #RADIANS
	* @see #DEGREE
	*/
	public final static int GRAD = 3;

	/** Radians heading mode (unit of mesurement) alias.
	*
	* @see #RADIANS
	*/
	public final static int RAD = RADIANS;

	/** Degree heading mode (unit of mesurement) alias.
	*
	* @see #DEGREE
	*/
	public final static int DEG = DEGREE;

	// --------

	/** Graphics context to work on.
	*
	* @see #getGraphics
	*
	* <p>
	* m_graphics can be null, in which case the draw operations are no-ops.
	*/
	private Graphics m_graphics = null;

	/** Bounding rectangle.
	*
	* <p>
	* m_bounds can be null, in which case the draw operations are no-ops.
	*/
	private Rectangle m_bounds = null;

	/** X scaling.
	*
	* <p>
	* Default is 1.0 (none).
	*/
	private double m_scaleX = 1.0;

	/** Y scaling.
	*
	* <p>
	* Default is 1.0 (none).
	*/
	private double m_scaleY = 1.0;

	/** X transforming.
	*
	* <p>
	* Default is 0.0 (none).
	*/
	private double m_transformX = 0.0;

	/** Y transforming.
	*
	* <p>
	* Default is 0.0 (none).
	*/
	private double m_transformY = 0.0;

	// The coordinates and the heading have to be
	// more exact than the underlying screen resolution
	// to omit intermediate rounding errors.

	/** X coordinate.
	*
	* @see #xcor
	*
	* <p>
	* Starting X coordinate is 0.0.
	*/
	private double m_currentX = 0.0;

	/** Y coordinate.
	*
	* @see #ycor
	*
	* <p>
	* Starting Y coordinate is 0.0.
	*/
	private double m_currentY = 0.0;

	/** Heading.
	*
	* @see #heading
	* @see #setheading
	* @see #m_headingMode
	*
	* <p>
	* Zero heading means in X direction (i.e. to the right).
	*
	* <p>
	* Starting heading is 0.0, i.e. to the right.
	*
	* <p>
	* Different heading modes (units of mesurement) are provided
	* in the public API (functions heading, left, right, setheading):
	* RADIANS (circle is 2 pi), DEGREE (circle is 360), GRAD (circle is 400).
	* Default heading mode is RADIANS.
	* The internal heading mode is always radians.
	*/
	private double m_heading = 0.0;

	/** Heading mode / unit of measurement.
	*
	* @see #setHeadingMode
	* @see #getHeadingMode
	*
	* <p>
	* Default is RADIANS.
	*/
	private int m_headingMode = RADIANS;

	/** Pen state.
	*
	* @see #penup
	* @see #pendown
	* @see #drawstate
	*
	* <p>
	* Default is 'down' (true).
	*/
	private boolean m_penDown = true;

	/** Arguments used by the argument-less function versions.
	*/
	private double[] implicitArguments = {0.0, 0.0,};

	// --------

	/** Return source information.
	* This includes version with date, author with email address,
	* copyright and disclaimer.
	*/
	public static String getInfo()
	{
		return m_version + "\n" +
			m_author + "\n" +
			m_copyright + "\n\n" +
			m_disclaimer + "\n";
	}

	/** Turtle: get a new Turtle.
	*
	* <p>
	* Scale and transformation are default (i.e. none).
	*/
	public Turtle(Graphics graphics, Rectangle bounds)
	{
		// allow null graphics here, e.g. for unit tests, etc
		m_graphics = graphics;

		// allow null bounds here too
		m_bounds = bounds;
	}

	/** Turtle: get a new Turtle.
	*
	* <p>
	* Transformation is default (i.e. none).
	*/
	public Turtle(
		Graphics graphics,
		Rectangle bounds,
		double scaleX,
		double scaleY)
	{
		this(graphics, bounds);

		m_scaleX = scaleX;
		m_scaleY = scaleY;
	}

	/** Turtle: get a new Turtle.
	*/
	public Turtle(
		Graphics graphics,
		Rectangle bounds,
		double scaleX,
		double scaleY,
		double transformX,
		double transformY)
	{
		this(graphics, bounds, scaleX, scaleY);

		m_transformX = transformX;
		m_transformY = transformY;
	}

	/** Get the Graphics object.
	*
	* @see #m_graphics
	*/
	public Graphics getGraphics()
	{
		return m_graphics;
	}

	// --------

	/** Set (put) pen up or down.
	*
	* @see #penup
	* @see #pendown
	*/
	protected Turtle setPen(boolean down)
	{
		m_penDown = down;

		return this;
	}

	/** Set and normalize heading.
	*/
	protected Turtle setAndNormalizeHeading(double phi)
	{
		m_heading = normalizeHeading(phi);

		return this;
	}

	/** Move pen relative.
	*
	* @see #penMoveTo
	*/
	protected Turtle penMoveToRelative(double dX, double dY)
	{
		return penMoveTo(
			m_currentX + dX,
			m_currentY + dY);
	}

	/** Move pen absolute.
	*
	* @see #penMoveToRelative
	*/
	protected Turtle penMoveTo(double x, double y)
	{
		// draw if the pen is down
		if (m_penDown) {
			draw(m_currentX, m_currentY, x, y);
		}

		// update Turtle cursor
		m_currentX = x;
		m_currentY = y;

		return this;
	}

	// -------- Logo programming language Turtle functions.

	/** Set (put) penstate in 'up' position.
	* [logo language turtle function]
	*
	* @see #pendown
	* @see #drawstate
	*/
	public Turtle penup()
	{
		return setPen(false);
	}

	/** Set (put) penstate in 'down' position.
	* [logo language turtle function]
	*
	* @see #penup
	* @see #drawstate
	*/
	public Turtle pendown()
	{
		return setPen(true);
	}

	/** Advance in current direction.
	* [logo language turtle function]
	*
	* @see #back
	*/
	public Turtle forward(double distance)
	{
		// note: can't have sin and cos in one call?

		return penMoveToRelative(
			distance * Math.cos(m_heading),
			distance * Math.sin(m_heading));
	}

	/** Back up in current direction.
	* [logo language turtle function]
	*
	* @see #forward
	*/
	public Turtle back(double distance)
	{
		return forward(-distance);
	}

	/** Change heading (direction) to the left
	* (counterclockwise, positive direction).
	* [logo language turtle function]
	*
	* @see #right
	*
	* <p>
	* omega can be rad, deg, or grad, depending on the heading mode.
	*/
	public Turtle left(double omega)
	{
		// note: use heading() call to get it in the appropriate heading mode
		return setheading(heading() + omega);
	}

	/** Change heading (direction) to the right
	* (clockwise, negative direction).
	* [logo language turtle function]
	*
	* @see #left
	*
	* <p>
	* omega can be rad, deg, or grad, depending on the heading mode.
	*/
	public Turtle right(double omega)
	{
		return left(-omega);
	}

	/** Move to a specified point.
	* [logo language turtle function]
	*
	* @see #setx
	* @see #sety
	*/
	public Turtle setxy(double x, double y)
	{
		return penMoveTo(x, y);
	}

	/** Move to a specified point (only X changes).
	* [logo language turtle function]
	*
	* @see #sety
	* @see #setxy
	*/
	public Turtle setx(double x)
	{
		return penMoveTo(x, m_currentY);
	}

	/** Move to a specified point (only Y changes).
	* [logo language turtle function]
	*
	* @see #setx
	* @see #setxy
	*/
	public Turtle sety(double y)
	{
		return penMoveTo(m_currentX, y);
	}

	/** Set the heading.
	* [logo language turtle function]
	*
	* @see #heading
	*
	* <p>
	* phi can be rad, deg, or grad, depending on the heading mode.
	*/
	public Turtle setheading(double phi)
	{
		return setAndNormalizeHeading(headingE2I(phi));
	}

	/** Set heading towards a point.
	* [logo language turtle function]
	*/
	public Turtle towards(double x, double y)
	{
		double dX = x - m_currentX;
		double dY = y - m_currentY;

		double phi = 0.0;

		// check if both zero.
		if (dX != 0.0 || dY != 0.0) {
			// Note: a mismatch with some manual page concerning
			// the atan2 argument order has been observed.
			phi = Math.atan2(dY, dX);
		}

		return setAndNormalizeHeading(phi);
	}

	/** Go home; heading to zero.
	* [logo language turtle function]
	*
	* @see #setxy
	*/
	public Turtle home()
	{
		return setxy(0.0, 0.0).setAndNormalizeHeading(0.0);
	}

	// Logo programming language Turtle functions that return other than Turtle.

	/** Get X coordinate.
	* [logo language turtle function]
	*
	* @see #ycor
	*/
	public double xcor()
	{
		return m_currentX;
	}

	/** Get Y coordinate.
	* [logo language turtle function]
	*
	* @see #xcor
	*/
	public double ycor()
	{
		return m_currentY;
	}

	/** Get heading.
	* [logo language turtle function]
	*
	* @see #setheading
	*
	* <p>
	* The mesurement unit depends on the heading mode and can be rad, deg, or grad.
	*/
	public double heading()
	{
		return headingI2E(m_heading);
	}

	/** Get pen state.
	* [logo language turtle function]
	*
	* <p>
	* Note: only the turtle relevant stuff given.
	*/
	public boolean drawstate()
	{
		return m_penDown;
	}

	// -------- Logo language Turtle function abbreviations.

	/** Abbreviation for penup.
	* [logo language turtle function abbreviation]
	*
	* @see #penup
	*/
	public Turtle pu()
	{
		return penup();
	}

	/** Abbreviation for pendown.
	* [logo language turtle function abbreviation]
	*
	* @see #pendown
	*/
	public Turtle pd()
	{
		return pendown();
	}

	/** Abbreviation for forward.
	* [logo language turtle function abbreviation]
	*
	* @see #forward
	*/
	public Turtle fd(double distance)
	{
		return forward(distance);
	}

	/** Abbreviation for back.
	* [logo language turtle function abbreviation]
	*
	* @see #back
	*/
	public Turtle bk(double distance)
	{
		return back(distance);
	}

	/** Abbreviation for left.
	* [logo language turtle function abbreviation]
	*
	* @see #left
	*/
	public Turtle lt(double omega)
	{
		return left(omega);
	}

	/** Abbreviation for right.
	* [logo language turtle function abbreviation]
	*
	* @see #right
	*/
	public Turtle rt(double omega)
	{
		return right(omega);
	}

	/** Abbreviation for setheading.
	* [logo language turtle function abbreviation]
	*
	* @see #setheading
	*/
	public Turtle seth(double phi)
	{
		return setheading(phi);
	}

	// -------- Argument-less versions of the Logo programming language Turtle functions.

	/** Advance in current direction.
	*
	* @see #back
	*
	* <p>
	* Requires distance to be set by a call to setArgument().
	*/
	public Turtle forward()
	{
		return forward(getArgument());
	}

	/** Back up in current direction.
	*
	* @see #forward
	*
	* <p>
	* Requires distance to be set by a call to setArgument().
	*/
	public Turtle back()
	{
		return back(getArgument());
	}

	/** Change heading (direction) to the left
	* (counterclockwise, positive direction).
	*
	* @see #right
	*
	* <p>
	* Requires omega to be set by a call to setArgument().
	*/
	public Turtle left()
	{
		return left(getArgument());
	}

	/** Change heading (direction) to the right
	* (clockwise, negative direction).
	*
	* @see #left
	*
	* <p>
	* Requires omega to be set by a call to setArgument().
	*/
	public Turtle right()
	{
		return right(getArgument());
	}

	/** Move to a specified point.
	*
	* @see #setx
	* @see #sety
	*
	* <p>
	* Requires X and Y to be set by calls to setArgument().
	*/
	public Turtle setxy()
	{
		// get (pop) y, then x
		double y = getArgument();
		double x = getArgument();

		return setxy(x, y);
	}

	/** Move to a specified point (only X changes).
	*
	* @see #sety
	* @see #setxy
	*
	* <p>
	* Requires X to be set by a call to setArgument().
	*/
	public Turtle setx()
	{
		return setx(getArgument());
	}

	/** Move to a specified point (only Y changes).
	*
	* @see #setx
	* @see #setxy
	*
	* <p>
	* Requires Y to be set by a call to setArgument().
	*/
	public Turtle sety()
	{
		return sety(getArgument());
	}

	/** Set the heading.
	*
	* @see #heading
	*
	* <p>
	* Requires phi to be set by a call to setArgument().
	*/
	public Turtle setheading()
	{
		return setheading(getArgument());
	}

	/** Set heading towards a point.
	*
	* <p>
	* Requires X and Y to be set by calls to setArgument().
	*/
	public Turtle towards()
	{
		// get (pop) y, then x
		double y = getArgument();
		double x = getArgument();

		return towards(x, y);
	}

	// -------- Argument-less versions of the Logo language Turtle function abbreviations.

	/** Abbreviation for forward.
	*
	* @see #forward
	*/
	public Turtle fd()
	{
		return forward();
	}

	/** Abbreviation for back.
	*
	* @see #back
	*/
	public Turtle bk()
	{
		return back();
	}

	/** Abbreviation for left.
	*
	* @see #left
	*/
	public Turtle lt()
	{
		return left();
	}

	/** Abbreviation for right.
	*
	* @see #right
	*/
	public Turtle rt()
	{
		return right();
	}

	/** Abbreviation for setheading.
	*
	* @see #setheading
	*/
	public Turtle seth()
	{
		return setheading();
	}

	// -------- Java style naming for the Turtle functions.

	/** Java style name for penup.
	*
	* @see #penup
	*/
	public Turtle penUp()
	{
		return penup();
	}

	/** Java style name for pendown.
	*
	* @see #pendown
	*/
	public Turtle penDown()
	{
		return pendown();
	}

	/** Java style name for setxy.
	*
	* @see #setxy
	*/
	public Turtle setXY(double x, double y)
	{
		return setxy(x, y);
	}

	/** Java style name for setxy.
	*
	* @see #setxy
	*/
	public Turtle setXY()
	{
		return setxy();
	}

	/** Java style name for setx.
	*
	* @see #setx
	*/
	public Turtle setX(double x)
	{
		return setx(x);
	}

	/** Java style name for setx.
	*
	* @see #setx
	*/
	public Turtle setX()
	{
		return setx();
	}

	/** Java style name for sety.
	*
	* @see #sety
	*/
	public Turtle setY(double y)
	{
		return sety(y);
	}

	/** Java style name for sety.
	*
	* @see #sety
	*/
	public Turtle setY()
	{
		return sety();
	}

	/** Java style name for setheading.
	*
	* @see #setheading
	*/
	public Turtle setHeading(double phi)
	{
		return setheading(phi);
	}

	/** Java style name for setheading.
	*
	* @see #setheading
	*/
	public Turtle setHeading()
	{
		return setheading();
	}

	/** Java style name for xcor.
	*
	* @see #xcor
	*/
	public double xCor()
	{
		return xcor();
	}

	/** Java style name for xcor.
	*
	* @see #xcor
	*/
	public double xCoordinate()
	{
		return xcor();
	}

	/** Java style name for xcor.
	*
	* @see #xcor
	*/
	public double getXCoordinate()
	{
		return xcor();
	}

	/** Java style name for ycor.
	*
	* @see #ycor
	*/
	public double yCor()
	{
		return ycor();
	}

	/** Java style name for ycor.
	*
	* @see #ycor
	*/
	public double yCoordinate()
	{
		return ycor();
	}

	/** Java style name for ycor.
	*
	* @see #ycor
	*/
	public double getYCoordinate()
	{
		return ycor();
	}

	/** Java style name for heading.
	*
	* @see #heading
	*/
	public double getHeading()
	{
		return heading();
	}

	/** Java style name for drawstate.
	*
	* @see #drawstate
	*/
	public boolean drawState()
	{
		return drawstate();
	}

	/** Java style name for drawstate.
	*
	* @see #drawstate
	*/
	public boolean getDrawState()
	{
		return drawstate();
	}

	// --------

	/** Set heading mode.
	*
	* @see #getHeadingMode
	* @see #RADIANS
	* @see #DEGREE
	* @see #GRAD
	*/
	public Turtle setHeadingMode(int mode)
	{
		switch (mode) {
		case RADIANS :
		case DEGREE :
		case GRAD :
			m_headingMode = mode;
			break;
		}

		// Note: does not throw on illegal mode,
		// just ignores it.

		return this;
	}

	/** Get heading mode.
	*
	* @see #setHeadingMode
	* @see #RADIANS
	* @see #DEGREE
	* @see #GRAD
	*/
	public int getHeadingMode()
	{
		return m_headingMode;
	}

	/** Set radians heading mode.
	*
	* @see #RADIANS
	*/
	public Turtle radians()
	{
		return setHeadingMode(RADIANS);
	}

	/** Set degree heading mode.
	*
	* @see #DEGREE
	*/
	public Turtle degree()
	{
		return setHeadingMode(DEGREE);
	}

	/** Set grad heading mode.
	*
	* @see #GRAD
	*/
	public Turtle grad()
	{
		return setHeadingMode(GRAD);
	}

	/** Abbreviation for radians.
	*
	* @see #radians
	*/
	public Turtle rad()
	{
		return radians();
	}

	/** Abbreviation for degree.
	*
	* @see #degree
	*/
	public Turtle deg()
	{
		return degree();
	}

	// --------

	/** Set an implicit argument.
	*
	* @see #getArgument
	*
	* <p>
	* Does not check for stack overflow,
	* earlier arguments just drop out.
	*/
	protected double setArgument(double argument)
	{
		// roll stack up
		for (int i = implicitArguments.length - 1; i > 0; i--) {
			implicitArguments[i] = implicitArguments[i - 1];
		}

		// push argument
		implicitArguments[0] = argument;

		return argument;
	}

	/** Get an implicit argument.
	*
	* @see #setArgument
	*
	* <p>
	* Does not check for stack underflow,
	* default arguments (0.0) are just returned.
	* This behavior typically makes the stack usage more tolerant.
	*/
	protected double getArgument()
	{
		// pop argument
		double argument = implicitArguments[0];

		// roll stack down
		for (int i = 0; i < implicitArguments.length - 1; i++) {
			implicitArguments[i] = implicitArguments[i + 1];
		}

		// fill with default (zero)
		implicitArguments[implicitArguments.length - 1] = 0.0;

		return argument;
	}

	/** Normalize heading.
	*/
	protected static double normalizeHeading(double phi)
	{
		// Note: denormalized typically by less than two circles,
		// so the loops won't take too long.

		double newHeading = phi;
		double twoPi = Math.PI * 2;

		// cut while larger or equal 2 pi.
		while (newHeading >= twoPi) {
			newHeading -= twoPi;
		}

		// increment while less zero.
		while (newHeading < 0.0) {
			newHeading += twoPi;
		}

		return newHeading;
	}

	/** Translate external heading mode to internal.
	*
	* @see #headingI2E
	*/
	protected double headingE2I(double phi)
	{
		switch (m_headingMode) {
		case RADIANS :
			return phi;
		case DEGREE :
			return phi * Math.PI / 180;
		case GRAD :
			return phi * Math.PI / 200;
		}
		return phi;
	}

	/** Translate internalheading mode to external.
	*
	* @see #headingE2I
	*/
	protected double headingI2E(double phi)
	{
		switch (m_headingMode) {
		case RADIANS :
			return phi;
		case DEGREE :
			return phi * 180 / Math.PI;
		case GRAD :
			return phi * 200 / Math.PI;
		}
		return phi;
	}

	/** Draw.
	*
	* @see #m_graphics
	* @see #m_bounds
	* @see #m_scaleX
	* @see #m_scaleY
	* @see #m_transformX
	* @see #m_transformY
	*
	* <p>
	* The logical Turtle layout is upside down in regards to the
	* java.awt.Graphics layout:
	*
	* <pre>
	* 0----------------> Graphics X
	* |
	* |        ^ Turtle Y
	* |        |
	* |        |
	* |        |
	* | -------0-------> Turtle X
	* |        |
	* |        |
	* |        |
	* |        |
	* |
	* V
	* Graphics Y
	* </pre>
	*
	* <p>
	* The coordinates are scaled and transformed and centered at the
	* middle of the supplied bounds:
	* <pre>
	* (int)(x * m_scaleX + m_transformX) + m_bounds.width / 2 + m_bounds.x
	* </pre>
	*/
	protected Graphics draw(double xFrom, double yFrom, double xTo, double yTo)
	{
		// check for null
		if (m_graphics == null || m_bounds == null) {
			return m_graphics;
		}

		m_graphics.drawLine(
			(int)(xFrom * m_scaleX + m_transformX) + m_bounds.width / 2 + m_bounds.x,
			(int)-(yFrom * m_scaleY + m_transformY) + m_bounds.height / 2 + m_bounds.y,
			(int)(xTo * m_scaleX + m_transformX) + m_bounds.width / 2 + m_bounds.x,
			(int)-(yTo * m_scaleY + m_transformY) + m_bounds.height / 2 + m_bounds.y);

		return m_graphics;
	}
}

